#######################################################################################################################
#######################################################################################################################
####### Continuous Time Meta Analysis (CoTiMA) - R script Version 1.4 (cf. Dormann, Guthier, & Cortina, 2019) #########
#######################################################################################################################
#######################################################################################################################


# NOTE: DO NOT EDIT/RUN THIS FILE. USE "CoTiMA Vx.x edit.R" INSTEAD. PUT BOTH FILES IN THE SAME WORKING DIRECTORY.

# contact: cdormann@uni-mainz.de

#######################################################################################################################
#######################################################################################################################
#######################################################################################################################
start.time <- Sys.time(); start.time

# (install and) activate required packages
if("MASS" %in% rownames(installed.packages()) == FALSE) {install.packages("MASS")} # for computing pseudo raw data 
if("foreach" %in% rownames(installed.packages()) == FALSE) {install.packages("foreach")} # for parallel processing (PP)
if("doParallel" %in% rownames(installed.packages()) == FALSE) {install.packages("doParallel")} # also for PP
if("ctsem" %in% rownames(installed.packages()) == FALSE) {   # if not already installed get ctsem 
  install.packages("ctsem")}
if("OpenMx" %in% rownames(installed.packages()) == FALSE) {   # if not already installed get NPSOL version of OpenMx
  source('https://openmx.ssri.psu.edu/software/getOpenMx.R')}

if ("ctsem" %in% (.packages())) detach("package:ctsem", unload=TRUE)
if ("OpenMx" %in% (.packages())) detach("package:OpenMx", unload=TRUE)
Sys.setenv(OMP_NUM_THREADS=parallel::detectCores()) # before library(OpenMx)
library(OpenMx)
mxOption(NULL, 'Number of Threads', parallel::detectCores())

library(foreach)
library(doParallel)
if(exists("unixLike") == FALSE) unixLike=FALSE 
if(unixLike == TRUE) registerDoParallel(detectCores()) 

if ("NPSOL" %in% mxAvailableOptimizers() == FALSE & NPSOL == TRUE) { # if not already installed but requested get NPSOL 
  source('https://openmx.ssri.psu.edu/software/getOpenMx.R')}        # ... version of OpenMx

library(MASS)  
library(ctsem)

# minor bug fixes
extraRefits <- c()

#######################################################################################################################

# create list with all study information
studyList <- list()
counter <- 1
while (is.numeric(tryCatch(get(paste("sampleSize", counter, sep = "")), error = function(e) e))) { 
  actualSampleNo <- counter 
  actualEmpcov <- get(paste("empcov", counter , sep = ""))
  actualDelta_t <- get(paste("delta_t", counter , sep = ""))
  actualSampleSize <- get(paste("sampleSize", counter , sep = ""))
  actualTimePoints <- length(actualDelta_t) + 1
  studyList[[counter ]] <- list(studyNumber=counter, empcov=actualEmpcov, delta_t=actualDelta_t,
                                sampleSize=actualSampleSize, originalStudyNo=actualSampleNo,
                                timePoints=actualTimePoints) 
  counter  <- counter + 1 
  if (isSymmetric(actualEmpcov) == FALSE) {
    cat("The correlation matrix of study no.", counter, "is not symmetric. Check and re-start!", "\n")
    stop("Good luck for the next try!")
  }
}
n.studies <- counter - 1; n.studies

### some stats
# sample size
allSampleSizes <-lapply(studyList, function(extract) extract$sampleSize); allSampleSizes
overallSampleSize <- sum(unlist(allSampleSizes)); overallSampleSize
meanSampleSize <- mean(unlist(allSampleSizes)); meanSampleSize
maxSampleSize <- max(unlist(allSampleSizes)); maxSampleSize
minSampleSize <- min(unlist(allSampleSizes)); minSampleSize
# lags
allDeltas <-lapply(studyList, function(extract) extract$delta_t); allDeltas
meanDelta <- mean(unlist(allDeltas)); meanDelta
maxDelta <- max(unlist(allDeltas)); maxDelta
minDelta <- min(unlist(allDeltas)); minDelta
# time points
allTpoints <-lapply(studyList, function(extract) extract$timePoints); allTpoints
overallTpoints <- sum(unlist(allTpoints)); overallTpoints
meanTpoints  <- mean(unlist(allTpoints)); meanTpoints
maxTpoints  <- max(unlist(allTpoints)); maxTpoints
minTpoints  <- min(unlist(allTpoints)); minTpoints

### create pseudo raw data for all studies
empraw <- lags <- list()
for (i in 1:n.studies) {
  currentVarnames <- c()
  currentSampleSize <- (lapply(studyList, function(extract) extract$sampleSize))[[i]]; currentSampleSize
  currentTpoints <- (lapply(studyList, function(extract) extract$timePoints))[[i]]; currentTpoints
  currentEmpcov <- (lapply(studyList, function(extract) extract$empcov))[[i]]; currentEmpcov
  currentLags <- (lapply(studyList, function(extract) extract$delta_t))[[i]]; currentLags
  for (j in 1:(currentTpoints)) {
    for (h in 1:nlatents) {
      currentVarnames <- c(currentVarnames, paste0("V",h,"_T", (j-1)))
      }
    }
  empraw[[i]] <- mvrnorm(n=currentSampleSize, mu=matrix(0, nrow=1, ncol=currentTpoints*nlatents), Sigma = currentEmpcov, empirical = TRUE)
  lags[[i]] <- matrix(currentLags, nrow=currentSampleSize, ncol=currentTpoints-1, byrow=TRUE)
  empraw[[i]] <- cbind(empraw[[i]], lags[[i]]); 
  colnames(empraw[[i]]) <- c(c(currentVarnames, paste0("dT", seq(1:(currentTpoints-1))))) 
  }

### create list of ctsem templates for studies with different numbers of time points
manifestNames <- c()
for (i in 1:nlatents) manifestNames <- c(manifestNames, paste0("V",i))
driftNames <- c()
for (i in 1:(nlatents)) {
  for (j in 1:(nlatents)) {
    driftNames <- c(driftNames, paste0("V",i,"toV", j))
    }
  }

# general ctsem model template
ctsemModelTemplate <- ctModel(n.latent=nlatents, n.manifest=nlatents, Tpoints=2, manifestNames=manifestNames,    # 2 waves in the template only
                              DRIFT=matrix(driftNames, nrow=nlatents, ncol=nlatents),
                              LAMBDA=diag(nlatents),
                              T0MEANS = matrix(c(0), nrow = nlatents, ncol = 1),
                              MANIFESTMEANS = matrix(c(0), nrow = nlatents, ncol = 1),
                              MANIFESTVAR=matrix(0, nrow=nlatents, ncol=nlatents))

# ctsem models for each primary study with the correct number of time points
ctsemModel <- list()
counter <- 1
for (i in unique(unlist(allTpoints))) {
  helper <- ctsemModelTemplate
  helper$Tpoints <- i
  ctsemModel[[counter]] <- helper
  counter <- counter +1
  }

### fit ctsem model to each primary study 
studyFit <- studyFitCI <- studyFit_Minus2LogLikelihood <- studyFit_estimatedParameters <- list() 
study_Drift_Coef <- study_Drift_SE <- list()
# loop through all primary studies
origRefits <- refits; origRefits
for (i in 1:n.studies) {  
  if (i %in% extraRefits) refits <- factorRefits * refits
  if (!(i %in% extraRefits)) refits <- origRefits
  # select correct template
  currentTpoints <- (lapply(studyList, function(extract) extract$timePoints))[[i]]; currentTpoints
  modelToSelect <- which(unique(allTpoints) == currentTpoints); modelToSelect
  currentModel <- ctsemModel[[modelToSelect]]; currentModel
  # Refit "refit" times 
  if (unixLike == TRUE) results <- foreach(h=1:refits) %dopar% {
    currentModel_fit <- list()
    set.seed(h)
    currentModel_fit[[h]] <- ctRefineTo(dat=empraw[[i]], currentModel, retryattempts=retryattempts)
    return(currentModel_fit[[h]]) 
  }
  if (unixLike != TRUE) results <- foreach(h=1:refits) %do% {
    currentModel_fit <- list()
    set.seed(h)
    currentModel_fit[[h]] <- ctRefineTo(dat=empraw[[i]], currentModel, retryattempts=retryattempts)
    return(currentModel_fit[[h]]) 
  }
  # Extract estimates & statistics
  allMinus2LogLikelihood <-lapply(results, function(extract) extract$mxobj$output$Minus2LogLikelihood); allMinus2LogLikelihood
  studyFit[[i]] <- results[[min(which(allMinus2LogLikelihood==min(unlist(allMinus2LogLikelihood))))]]
  studyFit_Minus2LogLikelihood[[i]] <- (summary(studyFit[[i]]))$omxsummary$Minus2LogLikelihood
  studyFit_estimatedParameters[[i]] <- (summary(studyFit[[i]]))$omxsummary$estimatedParameters
  study_Drift_Coef[[i]] <- (summary(studyFit[[i]]))$ctparameters[1:(nlatents^2), 1]
  study_Drift_SE[[i]] <- (summary(studyFit[[i]]))$ctparameters[1:(nlatents^2), 3]
  }


# Compute confidence intervals
if (confidenceIntervals == TRUE){ 
  if (unixLike != FALSE) { 
    results <- foreach(h=1:n.studies) %dopar% {
      studyFitCI[[h]] <- ctCI(studyFit[[h]], confidenceintervals = driftNames)
      return(studyFitCI[[h]]) }
  }
  if (unixLike == FALSE) {
    results <- foreach(h=1:n.studies) %do% {
      studyFitCI[[h]] <- ctCI(studyFit[[h]], confidenceintervals = driftNames)
      return(studyFitCI[[h]]) }
  }
  studyFitCI <- results; studyFitCI
}

# Combine summary information and fit statistics 
allStudies_Minus2LogLikelihood <- sum(unlist(studyFit_Minus2LogLikelihood)); allStudies_Minus2LogLikelihood
allStudies_estimatedParameters <- sum(unlist(studyFit_estimatedParameters)); allStudies_estimatedParameters
allStudies_df <- ((nlatents * unlist(allTpoints)) %*% ((nlatents * unlist(allTpoints)) +1 )) / 2 - 
  allStudies_estimatedParameters; allStudies_df
############################################# xxxx 
###allSampleSizes <-lapply(studyList, function(extract) extract$sampleSize); allSampleSizes
#if (confidenceIntervals == TRUE) allDriftCI <-lapply(studyFitCI, function(extract) summary(extract)$omxsummary$CI)
if (confidenceIntervals == TRUE) allDriftCI <-lapply(studyFitCI, function(extract) extract$mxobj$output$confidenceIntervals)
allStudiesDRIFT_effects <- matrix(t(cbind(unlist(study_Drift_Coef), unlist(study_Drift_SE)) ), n.studies, 2*nlatents^2, byrow=T)
# Label summary table
rownames(allStudiesDRIFT_effects) <- paste0("Study No ", 1:n.studies)
newColNames <- c()
for (j in 1:nlatents) {
  for (h in 1:nlatents) {
    newColNames <- c(newColNames, paste0("V",j,"toV", h), "(SE)")
  }
}
colnames(allStudiesDRIFT_effects) <- newColNames; round(allStudiesDRIFT_effects, digits)

#### Extract parameter estimates and paramter names to be used as start values later###############
est <- unlist(lapply(studyFit, function(extract) extract$mxobj$output$estimate)); est
startValues <- c(t(est)); startValues
noOfParams <- length(studyFit[[1]]$mxobj$output$estimate); noOfParams
headNames <- c(matrix(paste0(rep(c(paste0("Study_No_", 1:n.studies, "_")), noOfParams)), 
                      nrow=noOfParams, ncol=n.studies, byrow=T)); headNames
names(startValues) <- paste0(headNames, names(est)); startValues


#######################################################################################################################
############################################# CoTiMA (ctsem multigroup) ###############################################
#######################################################################################################################
# add missing variables (NA) and timelags (.00001) to datasets to achieve identical structure to study with most waves
empraw2 <- list()
for (i in 1:n.studies) {
  addTpoints <- maxTpoints - allTpoints[[i]]; addTpoints
  addVariables <- matrix(NA, nrow=allSampleSizes[[i]], ncol=(2*addTpoints)); dim(addVariables)
  addLags <- matrix(rep(.00001, addTpoints), nrow=allSampleSizes[[i]], ncol=addTpoints); dim(addLags)
  empraw2[[i]]  <- cbind(empraw[[i]][ , 1:(nlatents*allTpoints[[i]])], addVariables,
                         empraw[[i]][ , (nlatents*allTpoints[[i]]+1):(nlatents*allTpoints[[i]]+(allTpoints[[i]]-1))], addLags) }

# combine datasets
datawide_all <- matrix(NA, 0, ncol=dim(empraw2[[1]])[2]); datawide_all
for (i in 1:n.studies) datawide_all <- rbind(datawide_all, empraw2[[i]])
colnames(datawide_all) <- colnames(empraw[[min(which(max(unique(unlist(allTpoints)))==unlist(allTpoints)))]])

# create grouping variable
groups <- rep(c(paste0("Study_No_", 1:n.studies)), allSampleSizes)

######################################### Heterogeneity Model ########################################################

hetModel <- ctsemModel[[which(unique(unlist(allTpoints)) == max(unique(unlist(allTpoints))))]]

if (testHeterogeneityModel == TRUE) {
  if (exists("currentHetModel_fit")) rm(currentHetModel_fit)
  # Refit "refit" times
  if (unixLike == TRUE) results <- foreach(h=1:refits) %dopar% {
    currentHetModel_fit <- list()
    set.seed(h)
    currentHetModel_fit[[h]] <- ctMultigroupFit(dat = datawide_all, groupings = groups,
                                                omxStartValues=(startValues), retryattempts=retryattempts,
                                                ctmodelobj = hetModel)
    return(currentHetModel_fit[[h]]) 
  }
  if (unixLike != TRUE) results <- foreach(h=1:refits) %do% {
    currentHetModel_fit <- list()
    set.seed(h)
    currentHetModel_fit[[h]] <- ctMultigroupFit(dat = datawide_all, groupings = groups,
                                                omxStartValues=(startValues), retryattempts=retryattempts,
                                                ctmodelobj = hetModel)
    return(currentHetModel_fit[[h]]) 
  }
  # Select model with best fit
  allMinus2LogLikelihood <-lapply(results, function(extract) extract$mxobj$output$Minus2LogLikelihood); allMinus2LogLikelihood
  hetModelFit <- results[[min(which(allMinus2LogLikelihood==min(unlist(allMinus2LogLikelihood))))]]
  # Extract estimates & statistics
  allMinus2LogLikelihood <-lapply(results, function(extract) extract$mxobj$output$Minus2LogLikelihood); allMinus2LogLikelihood
  hetModelFit <- results[[min(which(allMinus2LogLikelihood==min(unlist(allMinus2LogLikelihood))))]]
  driftRows <- seq(1, nlatents^2, 1); driftRows
  for (i in 2:n.studies) driftRows <- c(driftRows, (i-1)*noOfParams + seq(1, nlatents^2, 1))
  driftRows
  hetModel_Drift_Coef <- hetModelFit$mxobj$output$estimate[driftRows]; hetModel_Drift_Coef
  hetModel_Drift_SE <- hetModelFit$mxobj$output$standardErrors[driftRows]; hetModel_Drift_SE
  hetModel_Minus2LogLikelihood <- hetModelFit$mxobj$output$Minus2LogLikelihood; hetModel_Minus2LogLikelihood
  hetModel_estimatedParameters <- length(hetModelFit$mxobj$output$estimate); hetModel_estimatedParameters
  hetModel_df <- ((nlatents * unlist(allTpoints)) %*% ((nlatents * unlist(allTpoints)) +1 )) / 2 - 
    hetModel_estimatedParameters; hetModel_df
  # Compute confidence intervals (may take ages)
  if (confidenceIntervals == TRUE) {
    hetModelFitCI <-  ctCI(hetModelFit, confidenceintervals = 'DRIFT')
    hetDriftCI <- hetModelFitCI$mxobj$output$confidenceIntervals
  }
  # Combine summary information and fit statistics 
  hetModelDRIFT_effects <- matrix(t(cbind((hetModel_Drift_Coef), (hetModel_Drift_SE)) ), n.studies, 2*nlatents^2, byrow=T)
  # Label summary table
  rownames(hetModelDRIFT_effects) <- paste0("Study No ", 1:n.studies)
  newColNames <- c()
  for (j in 1:nlatents) {
    for (h in 1:nlatents) {
          newColNames <- c(newColNames, paste0("V",j,"toV", h), "(SE)")
    }
  }
  colnames(hetModelDRIFT_effects) <- newColNames; hetModelDRIFT_effects
}

######################################### Homogeneity Models (CoTiMAs) ################################################

if (testDRIFTallModel == TRUE) {
  # relabel start values
  newHeadNames <- names(startValues); newHeadNames
  toBeDeleted <- grep("to", newHeadNames); toBeDeleted
  toBeDeleted <- toBeDeleted[-(1:nlatents^2)]; toBeDeleted
  newHeadNames <- newHeadNames[-toBeDeleted]; newHeadNames
  targetNames <- c()
  for (j in 1:nlatents) {
    for (h in 1:nlatents) {
      targetNames <- c(targetNames, paste0("V",j,"toV", h))
    }
  }
  newHeadNames[1:(nlatents^2)] <- targetNames; newHeadNames
  newStartValues <- startValues[-toBeDeleted]; newStartValues
  names(newStartValues) <- newHeadNames; newStartValues
  # copy model and constrain/fix drift parameters equal across primary studies
  homModel <- hetModel
  fixedModel <- homModel
  fixedModel$DRIFT <- matrix(rep("groupfixed", (nlatents^2)), nlatents, nlatents)
  # fit model
  if (exists("currentHomModel_fit")) rm(currentHomModel_fit)
  # Refit "refit" times
  if (unixLike == TRUE) results <- foreach(h=1:refits) %dopar% {
    currentHomModel_fit <- list()
    set.seed(h)
    currentHomModel_fit[[h]] <- ctMultigroupFit(dat=datawide_all, groupings = groups, retryattempts = retryattempts,
                                                omxStartValues=newStartValues,
                                                ctmodelobj = homModel, fixedmodel = fixedModel)
    return(currentHomModel_fit[[h]]) 
  }
  if (unixLike != TRUE) results <- foreach(h=1:refits) %do% {
    currentHomModel_fit <- list()
    set.seed(h)
    currentHomModel_fit[[h]] <- ctMultigroupFit(dat=datawide_all, groupings = groups, retryattempts = retryattempts,
                                                omxStartValues=newStartValues,
                                                ctmodelobj = homModel, fixedmodel = fixedModel)
    return(currentHomModel_fit[[h]]) 
  }
  # Select model with best fit
  homDRIFTallFit <- results[[min(which(allMinus2LogLikelihood==min(unlist(allMinus2LogLikelihood))))]]
  # Extract estimates & statistics
  homDRIFTallFit <- results[[min(which(allMinus2LogLikelihood==min(unlist(allMinus2LogLikelihood))))]]
  driftRows <- seq(1, nlatents^2, 1); driftRows
  homDRIFTall_Drift_Coef <- homDRIFTallFit$mxobj$output$estimate[driftRows]; homDRIFTall_Drift_Coef
  homDRIFTall_Drift_SE <- homDRIFTallFit$mxobj$output$standardErrors[driftRows]; homDRIFTall_Drift_SE
  Tvalue <- (homDRIFTall_Drift_Coef/homDRIFTall_Drift_SE); Tvalue  
  homDRIFTall_Minus2LogLikelihood  <- homDRIFTallFit$mxobj$output$Minus2LogLikelihood; homDRIFTall_Minus2LogLikelihood
  homDRIFTall_estimatedParameters  <- length(homDRIFTallFit$mxobj$output$estimate); homDRIFTall_estimatedParameters
  homDRIFTall_df <- ((nlatents * unlist(allTpoints)) %*% ((nlatents * unlist(allTpoints)) +1 )) / 2 - 
    homDRIFTall_estimatedParameters; homDRIFTall_df
  # Compute confidence intervals
  if (confidenceIntervals == TRUE) {
    homDRIFTallFitCI <-  ctCI(homDRIFTallFit, confidenceintervals = "DRIFT")
    homDRIFTallCI <- homDRIFTallFitCI$mxobj$output$confidenceIntervals
    }
  # Combine summary information
  homDRIFTallDRIFT_effects <- matrix(t(cbind((homDRIFTall_Drift_Coef), (homDRIFTall_Drift_SE), Tvalue)), 1, 3*length(driftRows), byrow=T)
  # Label summary table
  rownames(homDRIFTallDRIFT_effects) <- c("Fixed Effects")
  newColNames <- c()
  for (j in 1:nlatents) {
    for (h in 1:nlatents) {
      newColNames <- c(newColNames, paste0("V",j,"toV", h), "(SE)", "Tvalue")
    }
  }
  colnames(homDRIFTallDRIFT_effects) <- newColNames; homDRIFTallDRIFT_effects
}


### Test a series of homogeneity models with single drift effects invariant

if (testDRIFTSingleModel == TRUE) {
  
  homSingleModel <- fixedSingleModel <- homDRIFTSingleFitCI <- list()
  homDRIFTSingleFit <- list()
  homDRIFTSingle_Drift_Coef <- homDRIFTSingle_Drift_SE <- list()
  homDRIFTSingle_Minus2LogLikelihood <- homDRIFTSingle_estimatedParameters <- list()
  homDRIFTSingle_df  <- homDRIFTSingleCI <- list()
  homDRIFTSingleDRIFT_effects <- list()
  # vector of drift coefficients to be tested for invariance
  targetNames <- c()
  for (j in 1:nlatents) {
    for (h in 1:nlatents) {
      targetNames <- c(targetNames, paste0("V",j,"toV", h))
    }
  }
  # Loop through all drift coefficients to be tested for invariance
  for (i in 1:(nlatents^2)) {
    coeffNumber <- i 
    # relabel start values
    newHeadNames <- names(startValues); newHeadNames
    toBeDeleted <- grep(targetNames[coeffNumber], newHeadNames); toBeDeleted
    newHeadNames <- newHeadNames[-toBeDeleted[-1]]; newHeadNames
    newHeadNames[coeffNumber] <- targetNames[coeffNumber]; newHeadNames
    newStartValues <- startValues[-toBeDeleted[-1]]; newStartValues
    names(newStartValues) <- newHeadNames; newStartValues
    # copy model and constrain/fix drift parameters equal across primary studies
    homSingleModel[[i]] <- hetModel
    fixedSingleModel[[i]] <- homSingleModel[[i]]
    fixedSingleModel[[i]]$DRIFT[i] <- c("groupfixed")
    fixedSingleModel[[i]]
    # fit model
    if (exists("currentHomModel_fit")) rm(currentHomModel_fit)
    # Refit "refit" times
    if (unixLike == TRUE) results <- foreach(h=1:refits) %dopar% {
      currentHomModel_fit <- list()
      set.seed(h)
      currentHomModel_fit[[h]] <- ctMultigroupFit(dat= datawide_all, groupings = groups, retryattempts = retryattempts,
                                                  omxStartValues=newStartValues,
                                                  ctmodelobj = homSingleModel[[i]], fixedmodel = fixedSingleModel[[i]])
      return(currentHomModel_fit[[h]]) 
    }
    if (unixLike != TRUE) results <- foreach(h=1:refits) %do% {
      currentHomModel_fit <- list()
      set.seed(h)
      currentHomModel_fit[[h]] <- ctMultigroupFit(dat= datawide_all, groupings = groups, retryattempts = retryattempts,
                                                  omxStartValues=newStartValues,
                                                  ctmodelobj = homSingleModel[[i]], fixedmodel = fixedSingleModel[[i]])
      return(currentHomModel_fit[[h]]) 
    }
    # Select model with best fit
    homDRIFTSingleFit[[i]] <- results[[min(which(allMinus2LogLikelihood==min(unlist(allMinus2LogLikelihood))))]]
    # Extract estimates & statistics
    driftRows <- coeffNumber
    homDRIFTSingle_Drift_Coef[[i]] <- homDRIFTSingleFit[[i]]$mxobj$output$estimate; homDRIFTSingle_Drift_Coef[[i]]
    homDRIFTSingle_Drift_SE[[i]] <- homDRIFTSingleFit[[i]]$mxobj$output$standardErrors; homDRIFTSingle_Drift_SE[[i]]
    homDRIFTSingle_Minus2LogLikelihood[[i]] <- homDRIFTSingleFit[[i]]$mxobj$output$Minus2LogLikelihood; homDRIFTSingle_Minus2LogLikelihood[[i]] 
    homDRIFTSingle_estimatedParameters[[i]] <- length(homDRIFTSingleFit[[i]]$mxobj$output$estimate); homDRIFTSingle_estimatedParameters[[i]]
    homDRIFTSingle_df[[i]] <- ((nlatents * unlist(allTpoints)) %*% ((nlatents * unlist(allTpoints)) +1 )) / 2 - 
      homDRIFTSingle_estimatedParameters[[i]]; homDRIFTSingle_df[[i]]
    # Compute confidence intervals
    if (confidenceIntervals == TRUE) {
      homDRIFTSingleFitCI[[i]] <-  ctCI(homDRIFTSingleFit[[i]], confidenceintervals = 'DRIFT')
      homDRIFTSingleCI[[i]] <- homDRIFTSingleFitCI[[i]]$mxobj$output$confidenceIntervals
      }
    # Combine summary information
    homDRIFTSingleDRIFT_effects[[i]] <- matrix(t(cbind((homDRIFTSingle_Drift_Coef[[i]]), (homDRIFTSingle_Drift_SE[[i]])) ), 1, 2*length(driftRows), byrow=T)
    Tvalue <- (homDRIFTSingleDRIFT_effects[[i]][1]/homDRIFTSingleDRIFT_effects[[i]][2]); Tvalue
    # Label summary table
    newColNames <- c()
    for (j in 1:nlatents) {
      for (h in 1:nlatents) {
        newColNames <- c(newColNames, paste0("V",j,"toV", h), "(SE)")
      }
    }
    rownames(homDRIFTSingleDRIFT_effects[[i]]) <- c("Fixed Effects")
    colnames(homDRIFTSingleDRIFT_effects[[i]]) <- newColNames[((driftRows-1)*2+1):((driftRows-1)*2+2)]
    homDRIFTSingleDRIFT_effects[[i]] <- cbind(homDRIFTSingleDRIFT_effects[[i]], Tvalue); homDRIFTSingleDRIFT_effects[[i]]
  }
  }


###################################### Test Homogeneity of Cross or Auto Effects ######################################

if (testDRIFTCrossModel == TRUE & nlatents > 1) {
  
  ### fit equal cross-effects DRIFT homogeneity model
  
  ### Two Models are fitted: Model "2" with all cross effects homogenous across groups (as comparison model) ...
  ### and a model with all cross effects homogeneous AND INDENTICAL

  # The Model with all cross effects different but invariant 
  # relabel start values
  coeffNumber <- c()
  for (j in 1:nlatents) {
    for (h in 1:nlatents) {
      if (j != h) coeffNumber <- c(coeffNumber, ((j-1)*nlatents+h))
    }
  }
  newHeadNames <- names(startValues); newHeadNames
  toBeDeleted <- c()
  for (j in coeffNumber) toBeDeleted <- c(toBeDeleted, grep(targetNames[j], newHeadNames))
  toBeDeleted <- sort(toBeDeleted); toBeDeleted
  newHeadNames <- newHeadNames[-toBeDeleted[-(1:nlatents)]]; newHeadNames
  newStartValues <- startValues[-toBeDeleted[-(1:nlatents)]]; newStartValues
  names(newStartValues) <- newHeadNames; newStartValues
  # copy model and constrain/fix drift parameters equal across primary studies
  homCross2Model <- hetModel
  fixedCross2Model <- homCross2Model
  fixedCross2Model$DRIFT
  for (j in 1:nlatents) {
    for (h in 1:nlatents) {
      if (j != h) fixedCross2Model$DRIFT[j,h] <- c("groupfixed")
    }
  }
  # fit model
  if (exists("currentHomModel_fit")) rm(currentHomModel_fit)
  # Refit "refit" times
  if (unixLike == TRUE) results <- foreach(h=1:refits) %dopar% {
    currentHomModel_fit <- list()
    set.seed(h)
    currentHomModel_fit[[h]]  <- ctMultigroupFit(dat= datawide_all, groupings = groups, retryattempts = retryattempts,
                                                 omxStartValues=newStartValues,
                                                 ctmodelobj = homCross2Model, fixedmodel = fixedCross2Model)
    return(currentHomModel_fit[[h]]) 
  }
  if (unixLike != TRUE) results <- foreach(h=1:refits) %do% {
    currentHomModel_fit <- list()
    set.seed(h)
    currentHomModel_fit[[h]]  <- ctMultigroupFit(dat= datawide_all, groupings = groups, retryattempts = retryattempts,
                                                 omxStartValues=newStartValues,
                                                 ctmodelobj = homCross2Model, fixedmodel = fixedCross2Model)
    return(currentHomModel_fit[[h]]) 
  }
  # Select model with best fit
  homDRIFTCross2Fit <- results[[min(which(allMinus2LogLikelihood==min(unlist(allMinus2LogLikelihood))))]]
  # Extract estimates & statistics
  homDRIFTCross2_Minus2LogLikelihood <- homDRIFTCross2Fit$mxobj$output$Minus2LogLikelihood
  homDRIFTCross2_estimatedParameters <- length(homDRIFTCross2Fit$mxobj$output$estimate)
  homDRIFTCross2_df <- ((nlatents * unlist(allTpoints)) %*% ((nlatents * unlist(allTpoints)) +1 )) / 2 - 
    homDRIFTCross2_estimatedParameters; homDRIFTCross2_df

  # The Model with all cross effects equal and invariant
  toBeDeleted <- coeffNumber[2:(length(coeffNumber))]; toBeDeleted
  newStartValues <- newStartValues[-toBeDeleted]
  names(newStartValues)[2] <- "AllCrossEffects"
  # copy model and constrain/fix drift parameters equal across primary studies
  homCrossModel <- hetModel
  #targetNames <- homCrossModel$DRIFT; targetNames
  for (j in 1:nlatents) {
    for (h in 1:nlatents) {
      if (j != h) homCrossModel$DRIFT[h,j] <- "AllCrossEffects"
    }
  }
  #homCrossModel$DRIFT
  fixedCrossModel <- homCrossModel
  for (j in 1:nlatents) {
    for (h in 1:nlatents) {
      if (j != h) fixedCrossModel$DRIFT[j,h] <- c("groupfixed")
    }
  }
  # fit model
  if (exists("currentHomModel_fit")) rm(currentHomModel_fit)
  if (unixLike == TRUE) results <- foreach(h=1:refits) %dopar% {
    currentHomModel_fit <- list()
    set.seed(h)
    currentHomModel_fit[[h]] <- ctMultigroupFit(dat= datawide_all, groupings = groups, retryattempts = retryattempts,
                                                omxStartValues=newStartValues,
                                                ctmodelobj = homCrossModel, fixedmodel = fixedCrossModel)
    return(currentHomModel_fit[[h]]) 
  }
  if (unixLike != TRUE) results <- foreach(h=1:refits) %do% {
    currentHomModel_fit <- list()
    set.seed(h)
    currentHomModel_fit[[h]] <- ctMultigroupFit(dat= datawide_all, groupings = groups, retryattempts = retryattempts,
                                                omxStartValues=newStartValues,
                                                ctmodelobj = homCrossModel, fixedmodel = fixedCrossModel)
    return(currentHomModel_fit[[h]]) 
  }
  # Select Model with best fit
  homDRIFTCrossFit <- results[[min(which(allMinus2LogLikelihood==min(unlist(allMinus2LogLikelihood))))]]
  # Extract estimates & statistics
  homDRIFTCross_Minus2LogLikelihood <- homDRIFTCrossFit$mxobj$output$Minus2LogLikelihood
  homDRIFTCross_estimatedParameters <- length(homDRIFTCrossFit$mxobj$output$estimate)
  driftRows <- coeffNumber
  homDRIFTCross_Drift_Coef <- homDRIFTCrossFit$mxobj$output$estimate[min(driftRows)]
  homDRIFTCross_Drift_SE <- homDRIFTCrossFit$mxobj$output$standardErrors[min(driftRows)]
  homDRIFTCross_df <- ((nlatents * unlist(allTpoints)) %*% ((nlatents * unlist(allTpoints)) +1 )) / 2 - 
    homDRIFTCross_estimatedParameters; homDRIFTCross_df
  # Compute confidence intervals
  if (confidenceIntervals == TRUE) {
    homDRIFTCrossFitCI <-  ctCI(homDRIFTCrossFit, confidenceintervals = 'DRIFT')
    homDRIFTCrossCI <- homDRIFTCrossFitCI$mxobj$output$confidenceIntervals
  }
  # Combine summary information
  homDRIFTCrossDRIFT_effects <- matrix(t(cbind((homDRIFTCross_Drift_Coef), (homDRIFTCross_Drift_SE)) ), 1, 2*length(min(driftRows)), byrow=T)
  # Label summary table
  rownames(homDRIFTCrossDRIFT_effects) <- c("Fixed Effects")
  colnames(homDRIFTCrossDRIFT_effects) <- c("AllCrossEffects", "(SE)") 
  Tvalue <- (homDRIFTCrossDRIFT_effects[1]/homDRIFTCrossDRIFT_effects[2]); Tvalue 
  homDRIFTCrossDRIFT_effects <- cbind(homDRIFTCrossDRIFT_effects, Tvalue); homDRIFTCrossDRIFT_effects
  }


### fit equal auto-effects DRIFT homogeneity model

if (testDRIFTAutoModel == TRUE & nlatents > 1) {
  
  ### Two Models are fitted: Model "2" with AutoX homogenous AND AutoY homogenous (as comparison model) ...
  ### and a model with AutoX and AutoY INDENTICAL AND HOMOGENOUS
  
  # The Model with all auto effects different but invariant 
  # relabel start values
  coeffNumber <- c()
  for (j in 1:nlatents) {
    for (h in 1:nlatents) {
      if (j == h) coeffNumber <- c(coeffNumber, ((j-1)*nlatents+h))
    }
  }
  newHeadNames <- names(startValues); newHeadNames
  toBeDeleted <- c()
  for (j in coeffNumber) toBeDeleted <- c(toBeDeleted, grep(targetNames[j], newHeadNames))
  toBeDeleted <- sort(toBeDeleted); toBeDeleted
  newHeadNames <- newHeadNames[-toBeDeleted[-(1:nlatents)]]; newHeadNames
  newStartValues <- startValues[-toBeDeleted[-(1:nlatents)]]; newStartValues
  names(newStartValues) <- newHeadNames; newStartValues
  # copy model and constrain/fix drift parameters equal across primary studies
  homAuto2Model <- hetModel
  fixedAuto2Model <- homAuto2Model
  fixedAuto2Model$DRIFT
  for (j in 1:nlatents) {
    for (h in 1:nlatents) {
      if (j == h) fixedAuto2Model$DRIFT[j,h] <- c("groupfixed")
    }
  }
  # fit model
  if (exists("currentHomModel_fit")) rm(currentHomModel_fit)
  if (unixLike == TRUE) results <- foreach(h=1:refits) %dopar% {
    currentHomModel_fit <- list()
    set.seed(h)
    currentHomModel_fit[[h]] <- ctMultigroupFit(dat= datawide_all, groupings = groups, retryattempts = retryattempts,
                                                omxStartValues=newStartValues,
                                                ctmodelobj = homAuto2Model, fixedmodel = fixedAuto2Model)
    return(currentHomModel_fit[[h]])
  }
  if (unixLike != TRUE) results <- foreach(h=1:refits) %do% {
    currentHomModel_fit <- list()
    set.seed(h)
    currentHomModel_fit[[h]] <- ctMultigroupFit(dat= datawide_all, groupings = groups, retryattempts = retryattempts,
                                                omxStartValues=newStartValues,
                                                ctmodelobj = homAuto2Model, fixedmodel = fixedAuto2Model)
    return(currentHomModel_fit[[h]])
  }
  # Select Model with best fit
  homDRIFTAuto2Fit <- results[[min(which(allMinus2LogLikelihood==min(unlist(allMinus2LogLikelihood))))]]
  # Extract estimates & statistics
  homDRIFTAuto2_Minus2LogLikelihood <- homDRIFTAuto2Fit$mxobj$output$Minus2LogLikelihood 
  homDRIFTAuto2_estimatedParameters <- length(homDRIFTAuto2Fit$mxobj$output$estimate)
  homDRIFTAuto2_df <- ((nlatents * unlist(allTpoints)) %*% ((nlatents * unlist(allTpoints)) +1 )) / 2 - 
    homDRIFTAuto2_estimatedParameters; homDRIFTAuto2_df
  

  # The Model with all auto effects equal and invariant
  coeffNumber
  toBeDeleted <- coeffNumber[2:(length(coeffNumber))]; toBeDeleted
  newStartValues <- newStartValues[-toBeDeleted]
  names(newStartValues)[1] <- "AllAutoEffects"
  # copy model and constrain/fix drift parameters equal across primary studies
  homAutoModel <- hetModel
  #targetNames <- homCrossModel$DRIFT; targetNames
  for (j in 1:nlatents) {
    for (h in 1:nlatents) {
      if (j == h) homAutoModel$DRIFT[h,j] <- "AllAutoEffects"
    }
  }
  fixedAutoModel <- homAutoModel
  for (j in 1:nlatents) {
    for (h in 1:nlatents) {
      if (j == h) fixedAutoModel$DRIFT[j,h] <- c("groupfixed")
    }
  }
  # fit model
  if (exists("currentHomModel_fit")) rm(currentHomModel_fit)
  if (unixLike == TRUE) results <- foreach(h=1:refits) %dopar% {
    currentHomModel_fit <- list()
    set.seed(h)
    currentHomModel_fit[[h]] <- ctMultigroupFit(dat= datawide_all, groupings = groups, retryattempts = retryattempts,
                                                omxStartValues=newStartValues,
                                                ctmodelobj = homAutoModel, fixedmodel = fixedAutoModel)
    
    return(currentHomModel_fit[[h]])
  }
  if (unixLike != TRUE) results <- foreach(h=1:refits) %do% {
    currentHomModel_fit <- list()
    set.seed(h)
    currentHomModel_fit[[h]] <- ctMultigroupFit(dat= datawide_all, groupings = groups, retryattempts = retryattempts,
                                                omxStartValues=newStartValues,
                                                ctmodelobj = homAutoModel, fixedmodel = fixedAutoModel)
    
    return(currentHomModel_fit[[h]])
  }
  # Select model with best fit
  homDRIFTAutoFit <- results[[min(which(allMinus2LogLikelihood==min(unlist(allMinus2LogLikelihood))))]]
  # Extract estimates & statistics
  homDRIFTAuto_Minus2LogLikelihood <- homDRIFTAutoFit$mxobj$output$Minus2LogLikelihood
  homDRIFTAuto_estimatedParameters <- length(homDRIFTAutoFit$mxobj$output$estimate)
  driftRows <- coeffNumber
  homDRIFTAuto_Drift_Coef <- homDRIFTAutoFit$mxobj$output$estimate[min(driftRows)]
  homDRIFTAuto_Drift_SE <- homDRIFTAutoFit$mxobj$output$standardErrors[min(driftRows)]
  homDRIFTAuto_df <- ((nlatents * unlist(allTpoints)) %*% ((nlatents * unlist(allTpoints)) +1 )) / 2 - 
    homDRIFTAuto_estimatedParameters; homDRIFTAuto_df
  # Compute confidence intervals
  if (confidenceIntervals == TRUE) {
    homDRIFTAutoFitCI <-  ctCI(homDRIFTAutoFit, confidenceintervals = 'DRIFT')
    homDRIFTAutoCI <- homDRIFTAutoFitCI$mxobj$output$confidenceIntervals
  }
  # Combine summary information
  homDRIFTAutoDRIFT_effects <- matrix(t(cbind((homDRIFTAuto_Drift_Coef), (homDRIFTAuto_Drift_SE)) ), 1, 2*length(min(driftRows)), byrow=T)
  rownames(homDRIFTAutoDRIFT_effects) <- c("Fixed Effects")
  colnames(homDRIFTAutoDRIFT_effects) <- c("AutoEffects", "(SE)") 
  Tvalue <- (homDRIFTAutoDRIFT_effects[1]/homDRIFTAutoDRIFT_effects[2]); Tvalue 
  homDRIFTAutoDRIFT_effects <- cbind(homDRIFTAutoDRIFT_effects, Tvalue); homDRIFTAutoDRIFT_effects

}

#######################################################################################################################
################## Model Comparisons (-2LL Difference Test ~ Chi-square Difference Tests) #############################
#######################################################################################################################

if (testDRIFTallModel == TRUE) {
  allStudies_homDRIFTall_Minus2LogLikelihood <- homDRIFTall_Minus2LogLikelihood -
    allStudies_Minus2LogLikelihood; allStudies_homDRIFTall_Minus2LogLikelihood 
  allStudies_homDRIFTall_df <- homDRIFTall_df -
    allStudies_df; allStudies_homDRIFTall_df
  allStudies_homDRIFTall_prob <- 1-pchisq(abs(allStudies_homDRIFTall_Minus2LogLikelihood), 
                                          allStudies_homDRIFTall_df); allStudies_homDRIFTall_prob
}

if (testDRIFTSingleModel == TRUE) {
  allStudies_homDRIFTSingle_Minus2LogLikelihood <- allStudies_homDRIFTSingle_df <- allStudies_homDRIFTSingle_prob <- list()
  
  for (i in 1:(nlatents^2)) {
    allStudies_homDRIFTSingle_Minus2LogLikelihood[[i]] <- homDRIFTSingle_Minus2LogLikelihood[[i]] -
      allStudies_Minus2LogLikelihood; allStudies_homDRIFTSingle_Minus2LogLikelihood[[i]] 
    allStudies_homDRIFTSingle_df[[i]] <- homDRIFTSingle_df[[i]] -
      allStudies_df; allStudies_homDRIFTSingle_df[[i]]
    allStudies_homDRIFTSingle_prob[[i]] <- 1-pchisq(abs(allStudies_homDRIFTSingle_Minus2LogLikelihood[[i]]), 
                                                    allStudies_homDRIFTSingle_df[[i]]); allStudies_homDRIFTSingle_prob[[i]]
  }
}


if (testDRIFTCrossModel == TRUE & nlatents > 1) {
  # Special: Comparison of a Model including different (but invariant) cross effects (Model "2") with 
  #   a model including equal and invariant cross effects
  homDRIFTCross_homDRIFTCross2_Minus2LogLikelihood <- homDRIFTCross2_Minus2LogLikelihood -
    homDRIFTCross_Minus2LogLikelihood; homDRIFTCross_homDRIFTCross2_Minus2LogLikelihood 
  homDRIFTCross_homDRIFTCross2_df <- homDRIFTCross_df -
    homDRIFTCross2_df; homDRIFTCross_homDRIFTCross2_df
  homDRIFTCross_homDRIFTCross2_prob <- 1-pchisq(abs(homDRIFTCross_homDRIFTCross2_Minus2LogLikelihood), 
                                                homDRIFTCross_homDRIFTCross2_df); homDRIFTCross_homDRIFTCross2_prob
}

if (testDRIFTAutoModel == TRUE  & nlatents > 1) {
  # Special: Comparison of a Model including different (but invariant) auto effects (Model "2") with 
  #   a model including equal and invariant auto effects
  homDRIFTAuto_homDRIFTAuto2_Minus2LogLikelihood <- homDRIFTAuto2_Minus2LogLikelihood -
    homDRIFTAuto_Minus2LogLikelihood; homDRIFTAuto_homDRIFTAuto2_Minus2LogLikelihood 
  homDRIFTAuto_homDRIFTAuto2_df <- homDRIFTAuto2_df -
    homDRIFTAuto_df; homDRIFTAuto_homDRIFTAuto2_df
  homDRIFTAuto_homDRIFTAuto2_prob <- 1-pchisq(abs(homDRIFTAuto_homDRIFTAuto2_Minus2LogLikelihood), 
                                              abs(homDRIFTAuto_homDRIFTAuto2_df)); homDRIFTAuto_homDRIFTAuto2_prob
}


#######################################################################################################################
################################### Fixed & Random Effects Analyses ###################################################
#######################################################################################################################

if (fixedAndRandomEffects == TRUE) {
  
  ### Egger's test
  # The standard normal deviate (SND), defined as the odds ratio divided by its standard error, is regressed against the ...
  # ... estimate's precision, the latter being defined as the inverse of the standard error (regression equation: SND= a+ bxprecision). ...
  # ... As precision depends largely on sample size, small trials will be close to zero on the Ã axis. Small trials may produce an ...
  # ... odds ratio that differs from unity, but because the standard error will be large, the resulting standard normal deviate will ...
  # ... again be close to zero. Small trials will thus be close to zero on both axesâthat is, close to the origin. 
  # Conversely, large studies will produce precise estimates and, if the treatment is effective, also produce large standard normal ...
  # ... deviates. The points from a homogeneous set of trials, not distorted by selection bias, will thus scatter about a line that ...
  # ... runs through the origin at standard normal deviate zero (a=0), with the slope b indicating the size and direction of effect. ...
  # ... This situation corresponds to a symmetrical funnel plot.
  # If there is asymmetry, with smaller studies showing effects that differ systematically from larger studies, the regression line ...
  # ... will not run through the origin. The intercept a provides a measure of asymmetryâthe larger its deviation from zero the more ... 
  # ... pronounced the asymmetry. If the smaller studies show big protective effects, they will force the regression line below the ...
  # .... origin on the logarithmic scale. Negative values will therefore indicate that smaller studies show more pronounced beneficial ... 
  # ... effects than larger studies. In some situations (for example, if there are several small trials but only one larger study) ... 
  # ... power is gained by weighting the analysis by the inverse of the variance of the effect estimate. 
  # We performed both weighted and unweighted analyses and used the output from the analysis yielding the intercept with the larger ...
  # ... deviation from zero.
  #
  # NEGATIVE INTERCEPTS INDICATE THAT SMALLER STUDIES PRODUCED MORE PRONOUNCED BENEFICIAL EFFECTS!!! 
  #
  
  DRIFTCoeff <- matrix(unlist(study_Drift_Coef), n.studies, nlatents^2, byrow=TRUE); DRIFTCoeff
  targetNames <- c()
  for (j in 1:nlatents) {
    for (h in 1:nlatents) {
      targetNames <- c(targetNames, paste0("V",j,"toV", h))
    }
  }
  colnames(DRIFTCoeff) <- targetNames; DRIFTCoeff
  DRIFTSE <- matrix(unlist(study_Drift_SE), n.studies, nlatents^2, byrow=TRUE)
  targetNames <- c()
  for (j in 1:nlatents) {
    for (h in 1:nlatents) {
      targetNames <- c(targetNames, paste0("SE(V",j,"toV", h, ")"))
    }
  }
  colnames(DRIFTSE) <- targetNames; DRIFTSE
  
  DRIFTCoeffSND <- DRIFTCoeff / DRIFTSE; DRIFTCoeffSND
  DRIFTPrecision <- c(rep(1, nlatents^2))/(DRIFTSE^2); DRIFTPrecision
  colnames(DRIFTPrecision) <- colnames(DRIFTCoeffSND); DRIFTPrecision
  
  message1 <- "The pos. & sign. intercept indicates that LARGER studies produced larger effects"
  message2 <- "The neg. & sign. intercept indicates that SMALLER studies produced larger effects"
  tmp <- c()
  
  # Note: The absolute value of the DV is used to yield standard interpretations for the intercept
  eggerDrift <- list()
  for (j in 1:(nlatents^2)) {
    eggerDrift[[j]] <- lm(abs(DRIFTCoeffSND[,j])~DRIFTPrecision[,j])
    eggerDrift[[j]]$message <- "Egger's test ist not significant"
    if (summary(eggerDrift[[j]])$coefficients[1,1] > 0 & summary(eggerDrift[[j]])$coefficients[1,4] < .05) {
      eggerDrift[[j]]$message <- message1
    }
    if (summary(eggerDrift[[j]])$coefficients[1,1] < 0 & summary(eggerDrift[[j]])$coefficients[1,4] < .05) {
      eggerDrift[[j]]$message <- message2
    }
  } 

  FREAResults <- list()

  FREAResults[[1]] <- "############# Eggers Test for DRIFT Parameter Estimates  ###############################"
  FREACounter <- 1
  for (j in 1:(nlatents^2)) {
    FREACounter <- FREACounter + 1
    FREAResults[[FREACounter]] <- paste0("-------------------------------- Eggers Test for ", 
                                         colnames(DRIFTCoeff)[j], "--------------------------------")
    FREACounter <- FREACounter + 1
    FREAResults[[FREACounter]] <- eggerDrift[[j]]$message
    FREACounter <- FREACounter + 1
    FREAResults[[FREACounter]] <- summary(eggerDrift[[j]])
  }

  # unweighted means
  DriftMeans <- colMeans(DRIFTCoeff); DriftMeans
  
  # FIXED EFFECTS ANALYSIS: Traditional Meta Analytic "Mean" calculation (T scores) #########################
  # Sum of within weights  and weight * effect size
  T_DriftWeights <- colSums(DRIFTPrecision); T_DriftWeights
  T_DriftMeans <- colSums(DRIFTCoeff * DRIFTPrecision); T_DriftMeans
  names(T_DriftMeans) <- names(T_DriftWeights); T_DriftMeans
  # fixed effects results
  FixedEffect_Drift <- T_DriftMeans/T_DriftWeights; FixedEffect_Drift
  FixedEffect_DriftVariance <- 1/T_DriftWeights; FixedEffect_DriftVariance
  FixedEffect_DriftSE <- FixedEffect_DriftVariance^.5; FixedEffect_DriftSE
  FixedEffect_DriftUpperLimit <- FixedEffect_Drift + 1.96*FixedEffect_DriftSE; FixedEffect_DriftUpperLimit
  FixedEffect_DriftLowerLimit <- FixedEffect_Drift - 1.96*FixedEffect_DriftSE; FixedEffect_DriftLowerLimit
  FixedEffect_DriftZ <- FixedEffect_Drift/FixedEffect_DriftSE; FixedEffect_DriftZ
  FixedEffect_DriftProb <- round(1-pnorm(abs(FixedEffect_DriftZ), 
                                         mean=c(rep(0, (nlatents^2))), sd=c(rep(1, (nlatents^2))), log=F), digits=digits); FixedEffect_DriftProb
  Q_Drift <- colSums(DRIFTPrecision * DRIFTCoeff^2)- (colSums(DRIFTPrecision * DRIFTCoeff))^2 / colSums(DRIFTPrecision); Q_Drift
  H2_Drift <- Q_Drift/(n.studies-1); H2_Drift
  I2_Drift <- (H2_Drift-1)/H2_Drift*100; I2_Drift
  # tau square
  T2_DriftWeights <- colSums(DRIFTPrecision^2); T2_DriftWeights
  cDrift <- T_DriftWeights-T2_DriftWeights/T_DriftWeights; cDrift
  tau2Drift <- (Q_Drift-(n.studies-1))/cDrift; tau2Drift
  SElnHDrift <- c()
  SElnHDrift[] <- 0
  for (j in 1:(nlatents^2)) {
    if (Q_Drift[j] > n.studies) SElnHDrift[j] <- 1/2*(log(Q_Drift[j])-log(n.studies-1))/((2*Q_Drift[j])^.5-(2*(n.studies-1)-1)^.5)
    if (Q_Drift[j] <= n.studies) SElnHDrift[j] <-  (1/(2*(n.studies-2)) * (1 - 1/(3*(n.studies-2)^.5)) )^.5
  }
  
  H2DriftUpperLimit <- exp(log(H2_Drift) + 1.96*SElnHDrift); H2DriftUpperLimit
  H2DriftLowerLimit <- exp(log(H2_Drift) - 1.96*SElnHDrift); H2DriftLowerLimit
  L <- exp(0.5*log(Q_Drift/(n.studies-1))-1.96*SElnHDrift)
  U <- exp(0.5*log(Q_Drift/(n.studies-1))+1.96*SElnHDrift)
  I2DriftUpperLimit <- (U^2-1)/U^2 * 100; I2DriftUpperLimit
  I2DriftLowerLimit <- (L^2-1)/L^2 * 100; I2DriftLowerLimit
  
  MeanOfDriftValues <- DriftMeans
  fixedEffectDriftResults <- rbind(MeanOfDriftValues, FixedEffect_Drift, FixedEffect_DriftVariance, FixedEffect_DriftSE,
                                   FixedEffect_DriftUpperLimit, FixedEffect_DriftLowerLimit,
                                   FixedEffect_DriftZ, FixedEffect_DriftProb, tau2Drift, Q_Drift, H2_Drift,
                                   H2DriftUpperLimit, H2DriftLowerLimit, I2_Drift, 
                                   I2DriftUpperLimit, I2DriftLowerLimit)
  round(fixedEffectDriftResults, digits)
  

  # RANDOM EFFECTS ANALYSIS ###############################################################################
  
  # total variance weighting
  Ttot_DriftWeights <- 0
  Ttot_DriftMeans <- 0
  tau2DriftExtended <- do.call(rbind, replicate(n.studies, tau2Drift, simplify=FALSE))
  Ttot_DriftWeights <-colSums(1/ (DRIFTSE^2 + tau2DriftExtended)); Ttot_DriftWeights
  Ttot_DriftMeans <- colSums(DRIFTCoeff * 1/ (DRIFTSE^2 + tau2DriftExtended)); Ttot_DriftMeans
  # random effects results
  RandomEffecttot_Drift <- Ttot_DriftMeans/Ttot_DriftWeights; RandomEffecttot_Drift
  RandomEffecttot_DriftVariance <- 1/Ttot_DriftWeights; RandomEffecttot_DriftVariance
  RandomEffecttot_DriftSE <- RandomEffecttot_DriftVariance^.5; RandomEffecttot_DriftSE
  RandomEffecttot_DriftUpperLimit <- RandomEffecttot_Drift + 1.96*RandomEffecttot_DriftSE; RandomEffecttot_DriftUpperLimit
  RandomEffecttot_DriftLowerLimit <- RandomEffecttot_Drift - 1.96*RandomEffecttot_DriftSE; RandomEffecttot_DriftLowerLimit
  RandomEffecttot_DriftZ <- RandomEffecttot_Drift/RandomEffecttot_DriftSE; RandomEffecttot_DriftZ
  RandomEffecttot_DriftProb <- round(1-pnorm(abs(RandomEffecttot_DriftZ), 
                                             mean=c(rep(0, (nlatents^2))), sd=c(rep(1, (nlatents^2))), log=F), digits=digits); RandomEffecttot_DriftProb
  
  RandomEffectDriftResults <- rbind(RandomEffecttot_Drift, RandomEffecttot_DriftVariance, RandomEffecttot_DriftSE,
                                    RandomEffecttot_DriftUpperLimit, RandomEffecttot_DriftLowerLimit, 
                                    RandomEffecttot_DriftZ, RandomEffecttot_DriftProb)
  
  
  ### Funnelplots
  # Funnel plots, plots of the trials' effect estimates against sample size, may be useful to assess the ...
  # ... validity of meta-analyses. The funnel plot is based on the fact that precision in estimating the ...
  # ... underlying treatment effect will increase as the sample size of component studies increases. Results ...
  # ... from small studies will scatter widely at the bottom of the graph, with the spread narrowing among ...
  # ... larger studies. In the absence of bias the plot will resemble a symmetrical inverted funnel. ...
  # ... Conversely, if there is bias, funnel plots will often be skewed and asymmetrical.
  # In funnel plots, the estimates of effect size obtained in the studies are plotted against the sample size.
  
  stretch <- 1.2
  figureName <- c()
  driftPlots <- list()
  for (i in 1:(nlatents^2)) {
    # determine range of X- and Y-axes
    pairs <- cbind(DRIFTCoeff[,i], DRIFTSE[,i]); pairs
    minX <- min(DRIFTCoeff[,i]); minX 
    maxX <- max(DRIFTCoeff[,i]); maxX 
    maxAbsX <- max(abs(c(minX, maxX))); maxAbsX
    avgX <-FixedEffect_Drift[i]; avgX
    yMax <- stretch * max(DRIFTSE[,i]); yMax
    # Determine pseudo confidence intervals (http://www.metafor-project.org/doku.php/plots:funnel_plot_variations)
    # The pseudo-confidence region has bounds theta hat +- 1.96 *SE, with theta hat is the estimated effect or ... 
    # ... outcome based on the fixed-effects model and SE is the standard error value from the y-axis. 
    lowLeft <- FixedEffect_Drift[i] - 1.96 * yMax; lowLeft
    lowRight <- FixedEffect_Drift[i] + 1.96 * yMax; lowRight
    xMin <- 0 - max(abs(lowLeft), abs(lowRight), abs(DRIFTCoeff[,i])); xMin
    xMax <- 0 + max(abs(lowLeft), abs(lowRight), abs(DRIFTCoeff[,i])); xMax
    
    plot.new()
    plot(pairs, xlab=colnames(DRIFTCoeff)[i], ylab="Standard Error", xlim = c(xMin, xMax), ylim = c(yMax, 0))
    polygon(c(lowLeft, avgX, lowRight), c(stretch * yMax, 0, stretch * yMax), lty=3)
    abline(v=avgX, col="black", lwd=1.5, lty=1)
    
    figureContent <- colnames(DRIFTCoeff)[i]
    figureName[i] <- paste0(figureFilePrefix, "_",  "Funnel Plot for ", figureContent, ".png")
    dev.copy(png, figureName[i], width = 8, height = 8, units = 'in', res = 300)
    dev.off()
  }
  }


#######################################################################################################################
################################################## Plotting ###########################################################
#######################################################################################################################


if (plotCrossEffects == TRUE || plotAutoEffects == TRUE) {
  
  counterOnSet <- counter <- c()
  
  # names for plot titles and filenames
  for (i in 1:(nlatents)) {
    for (j in 1:(nlatents)) {
      driftNames <- c(driftNames, paste0("V",i,"toV", j))
    }
  }
  
  
  discreteTimeAutoPlots <- list()
  
  DriftForPlot <- array(dim=c(n.studies, nlatents, nlatents))
  for (h in 1:n.studies) DriftForPlot[h, 1:nlatents,1:nlatents] <- matrix(DRIFTCoeff[h, ], nlatents, nlatents)
  
  if (testDRIFTallModel == TRUE) DriftForPlot.D <- matrix(homDRIFTall_Drift_Coef, nlatents, nlatents)
  
  # Define function to compute discrete parameters using drift parameters and time-scaling factors
  # 1=empirical Drift Matrix, 2 TimeScale, 3&4 row&column of return element of the discrete drift matrix
  discreteDrift <-function(driftMatrix, timeScale, number) {   
    discreteDriftValue <- expm(timeScale %x% driftMatrix)
    discreteDriftValue[number] }
  
  startPoint <- 0
  endPoint <- 3*round(maxDelta+1)
  noOfSteps <- endPoint*3; noOfSteps #520
  plotPairs <- list()
  stepWidth <- (endPoint - startPoint)/noOfSteps
  
  plotPairs <- array(dim=c(n.studies, noOfSteps, 2+nlatents^2))
  plotPairs.D <- array(dim=c(noOfSteps, 2+nlatents^2))
  
  for (currentTimeScale in 0:noOfSteps){
    for (h in 1:n.studies) {
      timeValue <- currentTimeScale * stepWidth; timeValue
      plotPairs[h,currentTimeScale,1] <- currentTimeScale; plotPairs[h,currentTimeScale,1]
      plotPairs[h,currentTimeScale,2] <- timeValue; plotPairs[h,currentTimeScale,2]
      for (j in 1:(nlatents^2)) {
        plotPairs[h,currentTimeScale,(2+j)] <- discreteDrift((DriftForPlot[h,1:nlatents,1:nlatents]),timeValue, j)
      }
    }
    if (testDRIFTallModel == TRUE) {
      plotPairs.D[currentTimeScale, 1] <- currentTimeScale #index
      plotPairs.D[currentTimeScale, 2] <- timeValue #time value 
      for (j in 1:(nlatents^2)) {
        plotPairs.D[currentTimeScale,(2+j)] <- discreteDrift((DriftForPlot.D[1:nlatents,1:nlatents]),timeValue, j)
      }
    }
  }
  
  yMin <- round(min(c(min(plotPairs[ , , seq(3, (2+nlatents^2), 1)]), min(plotPairs.D[ , seq(3, (2+nlatents^2), 1)]))) -.05, 1); yMin
  yMax <- round(max(c(min(plotPairs[ , , seq(3, (2+nlatents^2), 1)]), max(plotPairs.D[ , seq(3, (2+nlatents^2), 1)]))) +.05, 1); yMax
  yStep <- abs(yMin-yMax)*10; yStep
  
  
  if (plotAutoEffects == TRUE) {
    
    figureFileNameAuto <- c()
    
    #### Auto-regressive Effects of X ###
    counter <- 0
    for (j in seq(1, nlatents^2, (nlatents+1))) {    # select the diagonal elements of the drift matrix
      counter <- counter + 1
      for (h in 1:n.studies) {
        plot(plotPairs[h, ,2], plotPairs[h, ,2+j], type="l", col="grey", lwd=1.5, 
             main=paste0("Auto-regressive Effects of V", counter),
             xlab="Time Lags", ylab ="Auto-regressive Beta",
             xlim = c(0,endPoint), ylim=c(yMin,yMax),
             xaxp=c(0,endPoint,round(endPoint/3)), 
             yaxp=c(yMin, yMax, yStep),
             xaxt='n', yaxt='n', ann=FALSE)
        par(new=T) }
      if (testDRIFTallModel == TRUE) {
        plot(plotPairs.D[,2], plotPairs.D[,2+j], type="l", lty=2, col="black", lwd=2.5,
             main=paste0("Auto-regressive Effects of V", counter),
             xlab="Time Lags", ylab ="Auto-regressive Beta",
             xlim = c(0,endPoint), ylim=c(yMin,yMax),
             xaxp=c(0,endPoint,round(endPoint/3)), 
             yaxp=c(yMin, yMax, yStep))
      }
      par(new=F)
      if (plotAutoEffects == TRUE) {
        
        figureFileNameAuto[counter] <- paste0(figureFilePrefix,"_CoTiMA results for auto-regressive effects of V", counter, ".png")
        dev.copy(png, figureFileNameAuto[counter], width = 8, height = 8, units = 'in', res = 300)
        dev.off()
      }
    }  
  }
  
  if (plotCrossEffects == TRUE & nlatents > 1) {
    
    figureFileNameCross <- c()
    counterOnSet <- counter; counterOnSet
    
    coeffSeq <- seq(1, nlatents^2, 1)[!(seq(1, nlatents^2, 1) %in% seq(1, nlatents^2, (nlatents+1)))]; coeffSeq
    counter <- 0
    for (j in coeffSeq) {    # select the non-diagonal elements of the drift matrix
      counter <- counter + 1
      for (h in 1:n.studies) {
        plot(plotPairs[h, ,2], plotPairs[h, ,2+j], type="l", col="grey", lwd=1.5,
             main=paste0("Cross-lagged Effects of ", driftNames[j]),
             xlab="Time Lags", ylab ="Cross-lagged Beta",
             xlim = c(0,endPoint), ylim=c(yMin,yMax),
             xaxp=c(0,endPoint,round(endPoint/3)), 
             yaxp=c(yMin, yMax, yStep),
             xaxt='n', yaxt='n', ann=FALSE)
        par(new=T)
      }
      if (testDRIFTallModel == TRUE) {
        plot(plotPairs.D[,2], plotPairs.D[,2+j], type="l", lty=2, col="black", lwd=2.5,
             main=paste0("Cross-lagged Effects of ", driftNames[j]),
             xlab="Time Lags", ylab ="Cross-lagged Beta",
             xlim = c(0,endPoint), ylim=c(yMin,yMax),
             xaxp=c(0,endPoint,round(endPoint/3)), 
             yaxp=c(yMin, yMax, yStep))
      }
      par(new=F)
      
      if (plotCrossEffects == TRUE) {
        figureFileNameCross[counter] <- paste0(figureFilePrefix,"_CoTiMA results for cross-lagged effects of ", driftNames[j], ".png")
        dev.copy(png, figureFileNameCross[counter], width = 8, height = 8, units = 'in', res = 300)
        dev.off()
      }
    }
  }
}


end.time <- Sys.time()
time.taken <- end.time - start.time
st <- paste0("Computation started at: ", start.time); st
et <- paste0("Computation ended at: ", end.time); et
tt <- paste0("Computation lasted: ", round(time.taken, digits)); tt



###########################################################################################################
###################################### SAVE RESULTS AND FIGURES ###########################################
###########################################################################################################

sink(file = resultsFileName, append = TRUE, type = c("output"), split = TRUE)

cat(paste("########################################################################################", "", sep="\n"))
cat(paste("########################################################################################", "", sep="\n"))
cat(paste("########### Continuous Time Meta Analysis (CoTiMA) - R script Version 1.0 ##############", "", sep="\n"))
cat(paste("#################### (cf. Dormann, Guthier, & Cortina, 2019) ###########################", "", sep="\n"))
cat(paste("########################################################################################", "", sep="\n"))
cat(paste("########################################################################################", "", sep="\n"))
cat(" ", "", sep="\n")
cat(st, sep="\n")
cat(et, sep="\n")
cat(tt, sep="\n")
cat(" ", "", sep="\n")
cat(paste("########################### contact: cdormann@uni-mainz.de #############################", "", sep="\n"))
cat(" ", "", sep="\n")
cat(" ", "", sep="\n")
cat(" ", "", sep="\n")
cat(paste("########################################################################################", "", sep="\n"))
cat(paste("------------------------------ Primary Study Statistics---------------------------------", "", sep="\n"))
cat(paste("########################################################################################", "", sep="\n"))
cat(" ", "", sep="\n")
cat(paste("-------------------------------- Sample Sizes ------------------------------------------", "", sep="\n"))
cat("Sample Sizes (k):       ", unlist(allSampleSizes))
cat(" ", "", sep="\n")
cat("Overall Sample Size (N):", overallSampleSize)
cat(" ", "", sep="\n")
cat(paste("---------------------------------- Time Lags -------------------------------------------", "", sep="\n"))
cat("All Time Lags (deltas): ", sort(unlist(allDeltas)))
cat(" ", "", sep="\n")
cat("Mean Time Lag (delta):  ", meanDelta)
cat(" ", "", sep="\n")
cat(paste("--------------------------------- Time Points ------------------------------------------", "", sep="\n"))
cat("All Time Points (Waves):", sort(unlist(allTpoints)))
cat(" ", "", sep="\n")
cat("Sum of all Time Points: ", overallTpoints)
cat(" ", "", sep="\n")
cat("Mean No. of Time Points:", round(meanTpoints, digits))
cat(" ", "", sep="\n")
cat(" ", "", sep="\n")
cat(paste("########################################################################################", "", sep="\n"))
cat(paste("--------------- All Studies Analyzed Separately (~ Heterogeneity Model) ----------------", "", sep="\n"))
cat(paste("########################################################################################", "", sep="\n"))
cat(" ", "", sep="\n")
cat(paste("-------------------------------- Coefficients ------------------------------------------", "", sep="\n"))
cat(" ", "", sep="\n")
cat("All Drift Coefficients and their Standard Errors (SE):")
cat(" ", "", sep="\n")
print(round(allStudiesDRIFT_effects, digits))
cat(" ", "", sep="\n")
if (confidenceIntervals == TRUE) {
  cat("All Drift Coefficients and their 95% Confidence Intervals (Not Necessarily Symmetric):")
  cat(" ", "", sep="\n")
  for (i in 1:n.studies) {print(paste0("Study No. ", i)); print(round(allDriftCI[[i]][, 1:3], 3))}
}
cat(" ", "", sep="\n")
cat(" ", "", sep="\n")
cat(paste("------------------------------- Fit Statistics -----------------------------------------", "", sep="\n"))
cat(" ", "", sep="\n")
cat("-2 Log Likelihood:")
cat(" ", "", sep="\n")
cat(round(allStudies_Minus2LogLikelihood, digits))
cat(" ", "", sep="\n")
cat("Overall Number of Estimated Parameters:")
cat(" ", "", sep="\n")
cat(round(allStudies_estimatedParameters, digits))
cat(" ", "", sep="\n")
cat("Degrees of Freedom (Standard SEM type. NOT OpenMx/CTSEM type):")
cat(" ", "", sep="\n")
cat(round(allStudies_df, digits))
cat(" ", "", sep="\n")
cat(" ", "", sep="\n")

if (testHeterogeneityModel == TRUE) {
  cat(paste("########################################################################################", "", sep="\n"))
  cat(paste("-------- (Actual Heterogeneity Model (Results Should be Identical to Those Above) ------", "", sep="\n"))
  cat(paste("########################################################################################", "", sep="\n"))
  cat(" ", "", sep="\n")
  cat(paste("-------------------------------- Coefficients ------------------------------------------", "", sep="\n"))
  cat(" ", "", sep="\n")
  cat("All Drift Coefficients and their Standard Errors (SE):")
  cat(" ", "", sep="\n")
  print(round(hetModelDRIFT_effects, digits))
  cat(" ", "", sep="\n")
  if (confidenceIntervals == TRUE) {
    cat("All Drift Coefficients and their 95% Confidence Intervals (Not Necessarily Symmetric):")
    cat(" ", "", sep="\n")
    print(round(hetDriftCI[, 1:3], digits))
  }
  cat(" ", "", sep="\n")
  cat(" ", "", sep="\n")
  cat(paste("------------------------------- Fit Statistics -----------------------------------------", "", sep="\n"))
  cat(" ", "", sep="\n")
  cat("-2 Log Likelihood:")
  cat(" ", "", sep="\n")
  cat(round(hetModel_Minus2LogLikelihood, digits))
  cat(" ", "", sep="\n")
  cat("Overall Number of Estimated Parameters:")
  cat(" ", "", sep="\n")
  cat(round(hetModel_estimatedParameters, digits))
  cat(" ", "", sep="\n")
  cat("Degrees of Freedom (Standard SEM type. NOT OpenMx/CTSEM type):")
  cat(" ", "", sep="\n")
  cat(round(hetModel_df, digits))
  cat(" ", "", sep="\n")
  cat(" ", "", sep="\n")
} 

if (testDRIFTallModel == TRUE) {
  cat(paste("########################################################################################", "", sep="\n"))
  cat(paste("---------- Multi-Sample Homogeneity Model (all Drift Coefficients Invariant) -----------", "", sep="\n"))
  cat(paste("########################################################################################", "", sep="\n"))
  cat(" ", "", sep="\n")
  cat(paste("-------------------------------- Coefficients ------------------------------------------", "", sep="\n"))
  cat(" ", "", sep="\n")
  cat("Synthesized/Aggregated Drift Coefficients and their Standard Errors (SE):")
  cat(" ", "", sep="\n")
  print(round(homDRIFTallDRIFT_effects, digits))
  cat(" ", "", sep="\n")
  if (confidenceIntervals == TRUE) {
    cat("Synthesized/Aggregated Drift Coefficients and their 95% Confidence Intervals (Not Necessarily Symmetric):")
    cat(" ", "", sep="\n")
    print(round(homDRIFTallCI [, 1:3], digits))
  }
  cat(" ", "", sep="\n")
  cat(" ", "", sep="\n")
  cat(paste("------------------------------- Fit Statistics -----------------------------------------", "", sep="\n"))
  cat(" ", "", sep="\n")
  cat("-2 Log Likelihood:")
  cat(" ", "", sep="\n")
  cat(round(homDRIFTall_Minus2LogLikelihood, digits))
  cat(" ", "", sep="\n")
  cat("Overall Number of Estimated Parameters:")
  cat(" ", "", sep="\n")
  cat(round(homDRIFTall_estimatedParameters, digits))
  cat(" ", "", sep="\n")
  cat("Degrees of Freedom (Standard SEM type. NOT OpenMx/CTSEM type):")
  cat(" ", "", sep="\n")
  cat(round(homDRIFTall_df, digits))
  cat(" ", "", sep="\n")
  cat("Comparison with Unconstrained Modell (All Samples Analyzed Separately (~Heterogeneity Model)):")
  cat(" ", "", sep="\n")
  cat("Delta(df):")
  cat(" ", "", sep="\n")
  cat(round(allStudies_homDRIFTall_df, digits))
  cat(" ", "", sep="\n")
  cat("Delta(-2LL) (= Delta(Chi-square)):")
  cat(" ", "", sep="\n")
  cat(round(allStudies_homDRIFTall_Minus2LogLikelihood, digits))
  cat(" ", "", sep="\n")
  cat("p-value (with double number of digits):")
  cat(" ", "", sep="\n")
  cat(round(allStudies_homDRIFTall_prob, 2*digits))
  cat(" ", "", sep="\n")
  cat("A sign. value (p < .05) indicates that entire process (i.e., the full Drift Matrix) varies among primary studies.")
  cat(" ", "", sep="\n")
  cat(" ", "", sep="\n")
}

if (testDRIFTSingleModel == TRUE) {
  targetNames <- c()
  for (j in 1:nlatents) {
    for (h in 1:nlatents) {
      targetNames <- c(targetNames, paste0("V",j,"toV", h))
    }
  }
  for (j in 1:length(targetNames)) {
    #j <- 1
    cat(paste("########################################################################################", "", sep="\n"))
    cat(paste("---------- Multi-Sample Homogeneity Model (only ", targetNames[j], " Coefficient Invariant) ----------", "", sep=""))
    cat(sep="\n")
    cat(paste("########################################################################################", "", sep="\n"))
    cat(" ", "", sep="\n")
    cat(paste("-------------------------------- Coefficients ------------------------------------------", "", sep="\n"))
    cat(" ", "", sep="\n")
    cat("Synthesized/Aggregated Drift Coefficients and their Standard Errors (SE):")
    cat(" ", "", sep="\n")
    print(round(homDRIFTSingleDRIFT_effects[[j]], digits))
    cat(" ", "", sep="\n")
    if (confidenceIntervals == TRUE) {
      cat("Synthesized/Aggregated Drift Coefficients and their 95% Confidence Intervals (Not Necessarily Symmetric):")
      cat(" ", "", sep="\n")
      toPrint <- homDRIFTSingleCI[[j]][1, 1:3]
      toPrint <- matrix(toPrint, nrow=1, dimnames=list("Fixed Effects", names(toPrint)))
      print(round(toPrint, digits))
    }
    cat(" ", "", sep="\n")
    cat(" ", "", sep="\n")
    cat(paste("------------------------------- Fit Statistics -----------------------------------------", "", sep="\n"))
    cat(" ", "", sep="\n")
    cat("-2 Log Likelihood:")
    cat(" ", "", sep="\n")
    cat(round(homDRIFTSingle_Minus2LogLikelihood[[j]], digits))
    cat(" ", "", sep="\n")
    cat("Overall Number of Estimated Parameters:")
    cat(" ", "", sep="\n")
    cat(round(homDRIFTSingle_estimatedParameters[[j]], digits))
    cat(" ", "", sep="\n")
    cat("Degrees of Freedom (Standard SEM type. NOT OpenMx/CTSEM type):")
    cat(" ", "", sep="\n")
    cat(round(homDRIFTSingle_df[[j]], digits))
    cat(" ", "", sep="\n")
    cat("Comparison with Unconstrained Modell (All Samples Analyzed Separately (~Heterogeneity Model)):")
    cat(" ", "", sep="\n")
    cat("Delta(df):")
    cat(" ", "", sep="\n")
    cat(round(allStudies_homDRIFTSingle_df[[j]], digits))
    j
    cat(" ", "", sep="\n")
    cat("Delta(-2LL) (= Delta(Chi-square)):")
    cat(" ", "", sep="\n")
    cat(round(allStudies_homDRIFTSingle_Minus2LogLikelihood[[j]], digits))
    cat(" ", "", sep="\n")
    cat("p-value (with double number of digits):")
    cat(" ", "", sep="\n")
    cat(round(allStudies_homDRIFTSingle_prob[[j]], 2*digits))
    cat(" ", "", sep="\n")
    cat("A sign. value (p < .05) indicates that the", targetNames[j] , "effects vary among primary studies.")
    cat(" ", "", sep="\n")
    cat(" ", "", sep="\n")
  }
}


if (testDRIFTCrossModel == TRUE & nlatents > 1) {
  cat(paste("########################################################################################", "", sep="\n"))
  cat(paste("----- Multi-Sample Homogeneity Model with (all Cross Effects Identical & Invariant) ----", "", sep="\n"))
  cat(paste("########################################################################################", "", sep="\n"))
  cat(" ", "", sep="\n")
  cat(paste("-------------------------------- Coefficients ------------------------------------------", "", sep="\n"))
  cat(" ", "", sep="\n")
  cat("Synthesized/Aggregated Drift Coefficients and their Standard Errors (SE):")
  cat(" ", "", sep="\n")
  print(round(homDRIFTCrossDRIFT_effects, digits))
  cat(" ", "", sep="\n")
  if (confidenceIntervals == TRUE) {
    cat("Synthesized/Aggregated Drift Coefficients and their 95% Confidence Intervals (Not Necessarily Symmetric):")
    cat(" ", "", sep="\n")
    toPrint <- homDRIFTCrossCI[which(rownames(homDRIFTCrossCI) == "AllCrossEffects"), 1:3]
    toPrint <- matrix(toPrint, nrow=1, dimnames=list("Fixed Effects", names(toPrint)))
    print(round(toPrint, digits))
    }
  cat(" ", "", sep="\n")
  cat(" ", "", sep="\n")
  cat(paste("------------------------------- Fit Statistics -----------------------------------------", "", sep="\n"))
  cat(" ", "", sep="\n")
  cat("-2 Log Likelihood:")
  cat(" ", "", sep="\n")
  cat(round(homDRIFTCross_Minus2LogLikelihood, digits))
  cat(" ", "", sep="\n")
  cat("Overall Number of Estimated Parameters:")
  cat(" ", "", sep="\n")
  cat(round(homDRIFTCross_estimatedParameters, digits))
  cat(" ", "", sep="\n")
  cat("Degrees of Freedom (Standard SEM type. NOT OpenMx/CTSEM type):")
  cat(" ", "", sep="\n")
  cat(round(homDRIFTCross_df, digits))
  cat(" ", "", sep="\n")
  cat("Comparison with Unconstrained Modell (See Note below)")
  cat(" ", "", sep="\n")
  cat("Delta(df):")
  cat(" ", "", sep="\n")
  cat(round(homDRIFTCross_homDRIFTCross2_df, digits))
  cat(" ", "", sep="\n")
  cat("Delta(-2LL) (= Delta(Chi-square)):")
  cat(" ", "", sep="\n")
  cat(round(abs(homDRIFTCross_homDRIFTCross2_Minus2LogLikelihood), digits))
  cat(" ", "", sep="\n")
  cat("p-value (with double number of digits):")
  cat(" ", "", sep="\n")
  cat(round(homDRIFTCross_homDRIFTCross2_prob, 2*digits))
  cat(" ", "", sep="\n")
  cat("Two models were fitted.")
  cat(" ", "", sep="\n")
  cat("A model (not reported) with every cross effect invariant across primary studies (but, e.g. V1toV2 and V2toV1 different).")
  cat(" ", "", sep="\n")
  cat("A second model (reported above) with all cross effects being equal.")
  cat(" ", "", sep="\n")
  cat("A sign. value (p < .05) indicates that the synthesized/aggregated cross effects are probably NOT equal.")
  cat(" ", "", sep="\n")
  cat(" ", "", sep="\n")
}

if (testDRIFTAutoModel == TRUE & nlatents > 1) {
  cat(paste("########################################################################################", "", sep="\n"))
  cat(paste("------ Multi-Sample Homogeneity Model with all Auto Effects Identical & Invariant ------", "", sep="\n"))
  cat(paste("########################################################################################", "", sep="\n"))
  cat(" ", "", sep="\n")
  cat(paste("-------------------------------- Coefficients ------------------------------------------", "", sep="\n"))
  cat(" ", "", sep="\n")
  cat("Synthesized/Aggregated Drift Coefficients and their Standard Errors (SE):")
  cat(" ", "", sep="\n")
  print(round(homDRIFTAutoDRIFT_effects, digits))
  cat(" ", "", sep="\n")
  if (confidenceIntervals == TRUE) {
    cat("Synthesized/Aggregated Drift Coefficients and their 95% Confidence Intervals (Not Necessarily Symmetric):")
    cat(" ", "", sep="\n")
    toPrint <- homDRIFTAutoCI [1, 1:3]
    toPrint <- matrix(toPrint, nrow=1, dimnames=list("Fixed Effects", names(toPrint)))
    print(round(toPrint, digits))
    }
  cat(" ", "", sep="\n")
  cat(" ", "", sep="\n")
  cat(paste("------------------------------- Fit Statistics -----------------------------------------", "", sep="\n"))
  cat(" ", "", sep="\n")
  cat("-2 Log Likelihood:")
  cat(" ", "", sep="\n")
  cat(round(homDRIFTAuto_Minus2LogLikelihood, digits))
  cat(" ", "", sep="\n")
  cat("Overall Number of Estimated Parameters:")
  cat(" ", "", sep="\n")
  cat(round(homDRIFTAuto_estimatedParameters, digits))
  cat(" ", "", sep="\n")
  cat("Degrees of Freedom (Standard SEM type. NOT OpenMx/CTSEM type):")
  cat(" ", "", sep="\n")
  cat(round(homDRIFTAuto_df, digits))
  cat(" ", "", sep="\n")
  cat("Comparison with Unconstrained Modell (See Note below)")
  cat(" ", "", sep="\n")
  cat("Delta(df):")
  cat(" ", "", sep="\n")
  cat(round(abs(homDRIFTAuto_homDRIFTAuto2_df), digits))
  cat(" ", "", sep="\n")
  cat("Delta(-2LL) (= Delta(Chi-square)):")
  cat(" ", "", sep="\n")
  cat(round(abs(homDRIFTAuto_homDRIFTAuto2_Minus2LogLikelihood), digits))
  cat(" ", "", sep="\n")
  cat("p-value (with double number of digits):")
  cat(" ", "", sep="\n")
  cat(round(homDRIFTAuto_homDRIFTAuto2_prob, 2*digits))
  cat(" ", "", sep="\n")
  cat("Two models were fitted.")
  cat(" ", "", sep="\n")
  cat("A model (not reported) with AutoX and AutoY invariant across primary studies (but AutoX and AutoY different).")
  cat(" ", "", sep="\n")
  cat("A second model (reported above) with AutoX and AutoY being equal.")
  cat(" ", "", sep="\n")
  cat("A sign. value (p < .05) indicates that the synthesized/aggregated Auto effects are probably NOT equal.")
  cat(" ", "", sep="\n")
  cat(" ", "", sep="\n")
}


if (fixedAndRandomEffects == TRUE) {
  cat(paste("########################################################################################", "", sep="\n"))
  cat(paste("---- Fixed Effects Analysis of Drift Coefficients (Taken From Heterogeneity Model) -----", "", sep="\n"))
  cat(paste("########################################################################################", "", sep="\n"))
  cat(" ", "", sep="\n")
  cat(paste("-------------------------------- Eggers Test -------------------------------------------", "", sep="\n"))
  cat(" ", "", sep="\n")
  cat("Note: If smaller studies show effects that differ systematically from larger studies, the regression line ")
  cat("will not run through the origin. A larger intercept indicates a larger asymmetry. ")
  cat("If the smaller studies show stronger (positive) effects, this will result in a regression line below the ")
  cat("origin. Negative intercepts thus indicate that smaller studies show stronger (positive) effects.") 
  cat(" ", "", sep="\n")
  print(FREAResults)
  
  cat(" ", "", sep="\n")
  cat("All Drift Effects in All Studies Analyzed Separately") 
  cat(" ", "", sep="\n")
  round(allStudiesDRIFT_effects, digits=digits)
  cat(" ", "", sep="\n")
  cat("Fixed Effects Analysis (Borenstein et al., 2007)") 
  cat(" ", "", sep="\n")
  fixedResults <-round(fixedEffectDriftResults, digits=digits)
  rownames(fixedResults) <- c("Mean", "Fixed Effect", "Variance of Fixed Effect", "Standard Error of Fixed Effect", 
                              "Upper Limit of Fixed Effect", "Lower Limit of Fixed Effect", "Z-Value", "p(Z)", 
                              "tau square", "Q", "H square (H2)", "Upper Limit of H2", "Lower Limit of H2", 
                              "I square (I2)", "Upper Limit of I2", "Lower Limit of I2")
  print(fixedResults)
  cat(" ", "", sep="\n")
  cat("Random Effects Analysis (Borenstein et al., 2007)") 
  cat(" ", "", sep="\n")
  randomResults <- round(RandomEffectDriftResults, digits=digits)
  cat(" ", "", sep="\n")
  rownames(randomResults) <- c("Random Effect", "Variance of Random Effect", "Standard Error of Random Effect", 
                               "Upper Limit of Random Effect", "Lower Limit of Random Effect", "Z-Value", "p(Z)")
  print(randomResults)
  cat(" ", "", sep="\n")
  cat("Note: NaN (not available) values could indicate negative between study variance (tau-squared).", "")
  cat("\"If the number of studies is very small, then it may be impossible to estimate the ...", "")
  cat("between-studies variance (tau-squared) with any precision. In this case, the fixed effect ...", "")
  cat("model may be the only viable option.\" (Borenstein, Hedges, & Rothstein, 2007, p 114).", "")
  cat(" ", "", sep="\n")
  cat(" ", "", sep="\n")
    cat(paste("-------------------------------- Funnel Plots  ----------------------------------------", "", sep="\n"))
  cat(" ", "", sep="\n")
  cat("Funnel Plots can be found in ", workingDirectory)
  cat(" ", "", sep="\n")
  cat("The files are labelled :")
  cat(" ", "", sep="\n")
  for (j in 1:length(figureName)) cat(figureName[j], "", sep="\n")
  cat(" ", "", sep="\n")
} 

if (plotCrossEffects == TRUE  & nlatents > 1) {
  
  cat(paste("########################################################################################", "", sep="\n"))
  cat(paste("------------------------ Discrete Time Plots of Cross-lagged Effects -------------------", "", sep="\n"))
  cat(paste("########################################################################################", "", sep="\n"))
  cat(" ", "", sep="\n")
  cat("Plots showing the discrete time cross-lagged effects can be found in ", workingDirectory)
  cat(" ", "", sep="\n")
  cat("The files are labelled :")
  cat(" ", "", sep="\n")
  for (j in 1:length(figureFileNameCross)) cat(figureFileNameCross[j], "", sep="\n")
  cat(" ", "", sep="\n")
  cat("Grey and solid lines represent the expected discrete time cross-lagged effects of the primary studies.")
  cat(" ", "", sep="\n")
  cat("The black and dashed line represents the CoTiMA-based expected discrete time cross-lagged effects.")
  cat(" ", "", sep="\n")
}

if (plotAutoEffects == TRUE) {
  
  cat(paste("########################################################################################", "", sep="\n"))
  cat(paste("----------------------- Discrete Time Plots of Autoregressive Effects ------------------", "", sep="\n"))
  cat(paste("########################################################################################", "", sep="\n"))
  cat(" ", "", sep="\n")
  cat("Plots showing the discrete time autoregressive effects can be found in ", workingDirectory)
  cat(" ", "", sep="\n")
  cat("The files are labelled :")
  cat(" ", "", sep="\n")
  for (j in 1:length(figureFileNameAuto)) cat(figureFileNameAuto[j], "", sep="\n")
  cat(" ", "", sep="\n")
  cat("Grey and solid lines represent the expected discrete time autoregressive effects of the primary studies.")
  cat(" ", "", sep="\n")
  cat("The black and dashed line represents the CoTiMA-based expected discrete time autoregressive effects.")
  cat(" ", "", sep="\n")
  
}
sink()

